@@ -4,6 +4,7 @@ import gzip
 import json
 import logging
 import multiprocessing
+import os
 import platform
 import shutil
 import sys
@@ -15,19 +16,22 @@ from typing import Any
 import httpx
 import pandas as pd
 import tqdm
+from tenacity import before_sleep, retry, retry_if_exception_type, wait_exponential
 from tqdm.contrib.concurrent import process_map
 from tqdm.contrib.logging import logging_redirect_tqdm
 
-from gpas import misc, __version__
+from gpas import __version__, misc
 from gpas.misc import (
     DEFAULT_ENVIRONMENT,
-    ENDPOINTS,
     ENVIRONMENTS,
+    ENVIRONMENTS_URLS,
     FILE_TYPES,
     GOOD_STATUSES,
 )
 from gpas.validation import build_validation_message, validate
 
+logger = logging.getLogger(__name__)
+
 
 def parse_token(token: Path) -> dict:
     return json.loads(Path(token).read_text())
@@ -50,11 +54,7 @@ def parse_mapping_csv(mapping_csv: Path) -> dict:
 
 def fetch_user_details(access_token, environment: ENVIRONMENTS) -> dict:
     """Test API authentication and fetch response from userOrgDtls endpoint"""
-    endpoint = (
-        ENDPOINTS[environment.value]["HOST"]
-        + ENDPOINTS[environment.value]["ORDS_PATH"]
-        + "userOrgDtls"
-    )
+    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['ORDS']}/userOrgDtls"
     try:
         logging.debug(f"Fetching user details {endpoint=}")
         r = httpx.get(endpoint, headers={"Authorization": f"Bearer {access_token}"})
@@ -121,14 +121,9 @@ async def fetch_status_async(
         "Authorization": f"Bearer {access_token}",
         "Content-Type": "application/json",
     }
-    endpoint = (
-        ENDPOINTS[environment.value]["HOST"]
-        + ENDPOINTS[environment.value]["API_PATH"]
-        + "get_sample_detail"
-    )
-
+    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['API']}/get_sample_detail"
     limits = httpx.Limits(
-        max_keepalive_connections=10, max_connections=20, keepalive_expiry=10
+        max_keepalive_connections=5, max_connections=10, keepalive_expiry=10
     )
     transport = httpx.AsyncHTTPTransport(limits=limits, retries=5)
     async with httpx.AsyncClient(transport=transport, timeout=30) as client:
@@ -155,6 +150,11 @@ async def fetch_status_async(
     return records
 
 
+@retry(
+    retry=retry_if_exception_type(httpx.HTTPError),
+    wait=wait_exponential(multiplier=1, min=1, max=8),
+    before_sleep=before_sleep.before_sleep_log(logger, 10),
+)
 async def fetch_status_single_async(
     client, guid, url, headers, warn=False, n_retries=5
 ):
@@ -165,15 +165,16 @@ async def fetch_status_single_async(
         result = dict(sample=guid, status=status)
         if warn and status not in GOOD_STATUSES:
             with logging_redirect_tqdm():
-                logging.warning(f"Skipping {guid} (status {status})")
+                logging.info(f"Skipping {guid} with status {status}")
     else:
         result = dict(sample=guid, status="Unknown")
-        with logging_redirect_tqdm():
-            logging.warning(f"HTTP {r.status_code} ({guid})")
         if r.status_code == 401:
             raise RuntimeError(
                 f"Authorisation failed (HTTP {r.status_code}). Invalid token?"
             )
+        else:
+            r.raise_for_status()  # Raises retryable httpx.HTTPError
+
     return result
 
 
@@ -189,11 +190,7 @@ async def download_async(
         "Authorization": f"Bearer {access_token}",
         "Content-Type": "application/json",
     }
-    endpoint = (
-        ENDPOINTS[environment.value]["HOST"]
-        + ENDPOINTS[environment.value]["API_PATH"]
-        + "get_output"
-    )
+    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['API']}/get_output"
 
     unrecognised_file_types = set(file_types) - {t.name for t in FILE_TYPES}
     if unrecognised_file_types:
@@ -202,7 +199,7 @@ async def download_async(
         logging.info(f"Fetching file types {file_types}")
 
     limits = httpx.Limits(
-        max_keepalive_connections=8, max_connections=16, keepalive_expiry=100
+        max_keepalive_connections=5, max_connections=10, keepalive_expiry=10
     )
     transport = httpx.AsyncHTTPTransport(limits=limits, retries=5)
     async with httpx.AsyncClient(transport=transport, timeout=120) as client:
@@ -233,6 +230,11 @@ async def download_async(
         ]
 
 
+@retry(
+    retry=retry_if_exception_type(httpx.HTTPError),
+    wait=wait_exponential(multiplier=1, min=1, max=8),
+    before_sleep=before_sleep.before_sleep_log(logger, 10),
+)
 async def download_single_async(
     client, guid, file_type, url, headers, out_dir, name=None, retries=5
 ):
@@ -258,9 +260,7 @@ async def download_single_async(
                 Path(f"{prefix}.{file_types_extensions[file_type]}"), guid, name
             )
     else:
-        result = dict(sample=guid, status="Unknown")
-        with logging_redirect_tqdm():
-            logging.warning(f"Skipping {guid}.{file_type} (HTTP {r.status_code})")
+        r.raise_for_status()  # Raises retryable httpx.HTTPError
 
 
 def fetch_status(
@@ -276,14 +276,10 @@ def fetch_status(
         "Authorization": f"Bearer {access_token}",
         "Content-Type": "application/json",
     }
-    endpoint = (
-        ENDPOINTS[environment.value]["HOST"]
-        + ENDPOINTS[environment.value]["API_PATH"]
-        + "get_sample_detail/"
-    )
+    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['API']}/get_sample_detail"
     records = []
     for guid in tqdm.tqdm(guids):
-        r = httpx.get(url=endpoint + guid, headers=headers)
+        r = httpx.get(url=f"{endpoint}/{guid}", headers=headers)
         if r.is_success:
             if raw:
                 records.append(r.json())
@@ -384,7 +380,7 @@ class Sample:
         )
         return command
 
-    def _get_convert_bam_cmd(self, paired=False) -> str:
+    def _get_convert_bam_cmd(self, paired=False) -> misc.LoggedShellCommand:
         prefix = Path(self.working_dir) / Path(self.sample_name)
         if not self.paired:
             cmd = f'"{self.samtools_path}" fastq -0 "{prefix.with_suffix(".fastq.gz")}" "{self.bam}"'
@@ -490,6 +486,17 @@ class Batch:
         self.save_reads = save_reads
         self.user_agent_name = user_agent_name
         self.user_agent_version = user_agent_version
+        self.client_info = {
+            "name": "gpas-cli",
+            "version": __version__,
+            "platform": platform.system(),
+            "python_version": platform.python_version(),
+        }
+        if self.user_agent_name or self.user_agent_version:
+            self.client_info["user_agent"] = {
+                "name": self.user_agent_name,
+                "version": self.user_agent_version,
+            }
         if self.token:
             auth_result = fetch_user_details(
                 self.token["access_token"], self.environment
@@ -510,7 +517,7 @@ class Batch:
             self.user = None
             self.organisation = None
             self.permitted_tags = []
-            self.headers = None
+            self.headers = {}
             self.date_mask = None
         logging.debug(f"{self.upload_csv=}")
         self.df, self.schema_name = validate(self.upload_csv, self.permitted_tags)
@@ -604,10 +611,9 @@ class Batch:
                 "samples": checksums,
             }
         }
+        logging.debug(f"_fetch_guids(): {payload=}")
         endpoint = (
-            ENDPOINTS[self.environment.value]["HOST"]
-            + ENDPOINTS[self.environment.value]["ORDS_PATH"]
-            + "createSampleGuids"
+            f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/createSampleGuids"
         )
         logging.debug(f"Fetching guids; {endpoint=}")
         r = httpx.post(url=endpoint, data=json.dumps(payload), headers=self.headers)
@@ -631,6 +637,18 @@ class Batch:
     def _rename_fastqs(self):
         """Rename decontaminated fastqs using server-side guids"""
         for s in self.samples:
+            # Workaround for ReadItAndKeep making fastas when encountering empty fastqs
+            if s.clean_fastq and not s.clean_fastq.exists():
+                s.clean_fastq = s.working_dir / Path(s.sample_name + ".reads.fasta.gz")
+            if s.clean_fastq1 and not s.clean_fastq1.exists():
+                s.clean_fastq1 = s.working_dir / Path(
+                    s.sample_name + ".reads_1.fasta.gz"
+                )
+            if s.clean_fastq2 and not s.clean_fastq2.exists():
+                s.clean_fastq2 = s.working_dir / Path(
+                    s.sample_name + ".reads_2.fasta.gz"
+                )
+
             s.clean_fastq = (
                 s.clean_fastq.rename(s.working_dir / Path(s.guid + ".reads.fastq.gz"))
                 if s.clean_fastq
@@ -670,7 +688,6 @@ class Batch:
         target_path = self.out_dir / Path(self.batch_guid + ".mapping.csv")
         df.to_csv(target_path, index=False)
         self.mapping_path = target_path
-        logging.info(f"Saved mapping CSV to {self.mapping_path}")
 
     def _fetch_par(self):
         """Private method that calls ORDS to get a Pre-Authenticated Request.
@@ -681,11 +698,7 @@ class Batch:
         -------
         par: str
         """
-        endpoint = (
-            ENDPOINTS[self.environment.value]["HOST"]
-            + ENDPOINTS[self.environment.value]["ORDS_PATH"]
-            + "pars"
-        )
+        endpoint = f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/pars"
         logging.debug(f"Fetching PAR; {endpoint=} {self.headers=}")
         r = httpx.get(url=endpoint, headers=self.headers)
         if not r.is_success:
@@ -730,26 +743,18 @@ class Batch:
         if self.json_messages:
             misc.print_progress_message_json(action="upload", status="started")
         uploads = self._get_uploads()
-        if self.processes == 1:
-            for upload in uploads:
-                misc.upload_sample(
-                    upload=upload,
-                    headers=self.headers,
-                    json_messages=self.json_messages,
-                )
-        else:
-            process_map(
-                partial(
-                    misc.upload_sample,
-                    headers=self.headers,
-                    json_messages=self.json_messages,
-                ),
-                uploads,
-                max_workers=10,
-                desc=f"Uploading {len(uploads)} sample(s) (10 connections)",
-                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
-                leave=False,
-            )
+        process_map(
+            partial(
+                misc.upload_sample,
+                headers=self.headers,
+                json_messages=self.json_messages,
+            ),
+            uploads,
+            max_workers=10,
+            desc=f"Uploading {len(uploads)} sample(s) (10 connections)",
+            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
+            leave=False,
+        )
         self.uploaded = True
         if self.json_messages:
             misc.print_progress_message_json(action="upload", status="finished")
@@ -757,30 +762,48 @@ class Batch:
             logging.info(f"Finished uploading {len(uploads)} sample(s)")
 
     def _finalise_submission(self):
+        @retry(
+            retry=retry_if_exception_type(httpx.HTTPError),
+            wait=wait_exponential(multiplier=1, min=60, max=120),
+            before_sleep=before_sleep.before_sleep_log(logger, 10),
+        )
+        def post_submission(submission: dict, headers: dict):
+            """Submit sample metadata to batches endpoint"""
+            endpoint = f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/batches"
+            logging.debug(f"post_submission(): {json.dumps(self.submission, indent=4)}")
+            r = httpx.post(
+                url=endpoint,
+                data=json.dumps(submission, ensure_ascii=False).encode("utf-8"),
+                headers=headers,
+                timeout=60,
+            )
+            logging.debug(f"post_submission(): {r.text=}")
+            r.raise_for_status()
+            if r.json().get("status") != "success":
+                raise misc.SubmissionError(r.json().get("errorMsg"))
+
+        @retry(
+            retry=retry_if_exception_type(httpx.HTTPError),
+            wait=wait_exponential(multiplier=1, min=10, max=40),
+            before_sleep=before_sleep.before_sleep_log(logger, 10),
+        )
+        def put_done_mark(batch_guid: str, headers: dict):
+            """Put upload done marker"""
+            url = self.par + batch_guid + "/upload_done.txt"
+            logging.debug(f"put_done_mark(): {url=}")
+            r = httpx.put(url=url, headers=self.headers)
+            logging.debug(f"put_done_mark(): {r.text=}")
+            r.raise_for_status()
+
         if not self.uploaded:
             raise RuntimeError("Reads not uploaded")
-        endpoint = (
-            ENDPOINTS[self.environment.value]["HOST"]
-            + ENDPOINTS[self.environment.value]["ORDS_PATH"]
-            + "batches"
-        )
 
         self.submission["batch"]["uploader"]["upload_finish"] = misc.oracle_timestamp()
-        logging.debug(json.dumps(self.submission, indent=4))
-        r = httpx.post(
-            url=endpoint,
-            data=json.dumps(self.submission, ensure_ascii=False).encode("utf-8"),
-            headers=self.headers,
+        post_submission(self.submission, self.headers)
+        put_done_mark(self.batch_guid, self.headers)
+        logging.info(
+            f"Submitted batch {self.batch_guid}, mapping CSV saved to {self.mapping_path}"
         )
-
-        r.raise_for_status()
-        logging.debug(f"POSTing JSON {r.text=}")
-        if r.json().get("status") != "success":
-            raise misc.SubmissionError(r.json().get("errorMsg"))
-        url = self.par + self.batch_guid + "/upload_done.txt"  # Finalisation mark
-        r = httpx.put(url=url, headers=self.headers)
-        r.raise_for_status()
-        logging.info(f"Finished uploading batch {self.batch_guid}")
         success_message = {
             "submission": {
                 "status": "success",
@@ -855,51 +878,64 @@ class Batch:
                 "run_numbers": self.run_numbers,
                 "samples": samples,
                 "uploader": {
-                    "name": "gpas-cli",
-                    "version": __version__,
-                    "platform": platform.system(),
-                    "python_version": platform.python_version(),
+                    **self.client_info,
                     "upload_start": self.uploaded_on,
                     "upload_finish": None,  # Set in _finalise_submission()
                 },
             },
         }
-        if self.user_agent_name or self.user_agent_version:
-            self.submission["batch"]["uploader"]["user_agent"] = {
-                "name": self.user_agent_name,
-                "version": self.user_agent_version,
+
+    def post_exception(self, exception: Exception):
+        """Report exceptions occurring during Batch.upload()"""
+        try:
+            if "PYTEST_CURRENT_TEST" in os.environ:  # Disable reporting under pytest
+                return
+            e_t, e_v, e_tb = misc.get_value_traceback_fmt(exception)
+            endpoint = (
+                f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/logUploaderError"
+            )
+            payload = {
+                "exception": {
+                    "class": e_t,
+                    "message": e_v,
+                    "traceback": e_tb,
+                    "timestamp": misc.oracle_timestamp(),
+                },
+                "uploader": self.client_info,
             }
+            logging.debug(f"Exception payload {payload=}")
+            r = httpx.post(url=endpoint, json=payload, headers=self.headers)
+            logging.debug(f"Exception submission {r.is_success=}")
+        except Exception:
+            pass
 
     def upload(self, dry_run: bool = False, save_reads: bool = False) -> None:
-        if self.processes > 1:
-            logging.info(f"Using {self.processes} processes")
-        self._decontaminate()
-        if not self.headers:
-            logging.warning("No token provided, quitting")
-            sys.exit()
-        self._hash_fastqs()
-        self._fetch_guids()
-        self._build_mapping_csv()
-        self._rename_fastqs()
-        if dry_run:
-            logging.info("Skipped submission")
-        else:
-            self._submit()
-
-        for s in self.samples:
-            logging.debug(f" {s.clean_fastq=}; {s.clean_fastq1=}; {s.clean_fastq2=}")
+        try:
+            if self.processes > 1:
+                logging.info(f"Using {self.processes} processes")
+            self._decontaminate()
+            if not self.headers:
+                logging.warning("No token provided, quitting")
+                sys.exit()
+            self._hash_fastqs()
+            self._fetch_guids()
+            self._build_mapping_csv()
+            self._rename_fastqs()
+            if dry_run:
+                logging.info("Skipped submission")
+            else:
+                self._submit()
+        except Exception as e:
+            self.post_exception(e)
+            raise e
 
     def _number_runs(self) -> None:
         """Enumerate unique values of run_number for submission"""
         samples_runs = self._get_sample_attrs("run_number")
-        runs = list(sorted(filter(None, samples_runs.values())))
-        if runs:  # More than just an empty string
-            runs_numbers = {r: i for i, r in enumerate(runs, start=1)}
-        else:
-            runs_numbers = {"": None}
+        samples_run_numbers = misc.number_runs(samples_runs)
         for s in self.samples:
-            s.gpas_run_number = runs_numbers[s.run_number]
-        self.run_numbers = list(filter(None, runs_numbers.values()))
+            s.gpas_run_number = samples_run_numbers[s.sample_name]
+        self.run_numbers = list(filter(None, samples_run_numbers.values()))
 
 
 def parse_decontamination_stats(stdout: str) -> dict:
@@ -912,8 +948,12 @@ def parse_decontamination_stats(stdout: str) -> dict:
     count_out = counts[2] + counts[3]
     delta = count_in - count_out
     assert delta >= 0
+    try:
+        fraction = round(delta / count_in, 4)
+    except ArithmeticError:  # ZeroDivisionError
+        fraction = 0
     return {
         "in": count_in,
         "out": count_out,
-        "fraction": round(delta / count_in, 4),
+        "fraction": fraction,
     }
